/*
Parker Corp Confidential 2016
Sensor Demo v2.01
10/15/2015
Last Rev Jan 29, 2016
*/

#define SPEED 1

#include "SPI.h"
#include "EEPROM.h"

// definitions 
#define OFF 0
#define ON  1
#define START  1
#define STOP  0
#define YES ON
#define NO OFF

//  names for Arduino's I/O-pins
const int SCLK          = 13;       // SPI-clock out
const int TWELVE_PIN    = 12;       // taken by SPI (could be used as input)
const int SDATA         = 11;       // SPI-data out
const int SDAC_SEL_n    = 10;       // low selects quad monitor-DAC
const int TPT1_PIN      =  9;       // test-point
const int EIGHT_PIN     =  8;       // (not used)

const int SEVEN_PIN     =  7;       // (not used)
const int CREF_IN       =  6;       // AIN+ pin to analog-comparator: switchable voltage
const int LED_ACT       =  5;       // Activity LED   
const int HI_LO_SEL     =  4;       // high selects low reference-in, low selects  high-reference
const int CHMUX_1       =  3;       // channel-enable mux inputs ...
const int CHMUX_0       =  2;       //    ...  00= CH_0, 01=CH_1, 10=CH_2, 11=ALL_OFF
const int UART_TX       =  1;       // used by UART
const int UART_RX       =  0;       // used by UART

const int POT_IN      =  A5;        // (potentiometer-input)
const int A4_PIN      =  A4;        // (not used: analog-input)
const int A5_PIN      =  A3;        // (not used: analog-input)
const int SNSR_CH0    =  A2;        // Sensor A input
const int SNSR_CH1    =  A1;        // Sensor B input
const int SNSR_CH2    =  A0;        // Sensor C input

// some I/O macros: these break portability (i.e., they are AVRxx8/Uno-specific) in
//  the cause of speed
#ifdef SPEED
#define TP2_ON       PORTD = PORTD |   128;     //  digitalWrite(SEVEN_PIN, ON)
#define TP2_OFF      PORTD = PORTD &  ~128;     //  digitalWrite(SEVEN_PIN, OFF)
#define HIGH_REF_ON PORTD = PORTD & ~16;       //  digitalWrite(HI_LO_SEL , OFF);
#define LOW_REF_ON  PORTD = PORTD |  16;       //  digitalWrite(HI_LO_SEL , ON);
#define EN_CH0      PORTD = (PORTD & ~(8+4));  //  digitalWrite(CHMUX_1, OFF); digitalWrite(CHMUX_0, OFF);
#define EN_CH1      PORTD = (PORTD & ~8) | 4; //  digitalWrite(CHMUX_1, OFF); digitalWrite(CHMUX_0, ON);
#define EN_CH2      PORTD = (PORTD & ~4) | 8; //  digitalWrite(CHMUX_1, ON); digitalWrite(CHMUX_0, ON);
#define ALL_CH_OFF  PORTD = (PORTD | (8+4) );
#define ACT_LED_ON   PORTD = PORTD |  32;    //  digitalWrite(LED_ACT, ON)
#define ACT_LED_TGL  PORTD = PORTD ^  32;    //  digitalWrite(LED_ACT, digitalRead(LED_ACT))
#define ACT_LED_OFF  PORTD = PORTD & ~32;    //  digitalWrite(LED_ACT, OFF)
#define SDAC_EN     PORTB = PORTB &  ~4;     //  digitalWrite(SDAC_SEL_n , OFF);
#define SDAC_DIS    PORTB = PORTB |   4;     //  digitalWrite(SDAC_SEL_n , ON);
#define TP1_ON      PORTB = PORTB |   2;     //  digitalWrite(TPT1_PIN, ON)
#define TP1_OFF     PORTB = PORTB &  ~2;     //  digitalWrite(TPT1_PIN, OFF)
#define TP1_TGL     PORTB = PORTB ^   2;     //  digitalWrite(TPT1_PIN, digitalRead(LED_ACT))

#else
#define ACT_LED_ON  digitalWrite(LED_ACT, ON)
#define ACT_LED_OFF  digitalWrite(LED_ACT, OFF)
#define ACT_LED_TGL  digitalWrite(LED_ACT, digitalRead(LED_ACT))
#define EN_CH0  digitalWrite(CHMUX_1, OFF); digitalWrite(CHMUX_0, OFF);
#define EN_CH1  digitalWrite(CHMUX_1, OFF); digitalWrite(CHMUX_0, ON);
#define EN_CH2  digitalWrite(CHMUX_1, ON); digitalWrite(CHMUX_0, OFF);
#define ALL_CH_OFF  digitalWrite(CHMUX_1, ON); digitalWrite(CHMUX_0, ON);
#define HIGH_REF_ON digitalWrite(HI_LO_SEL , OFF);
#define LOW_REF_ON digitalWrite(HI_LO_SEL , ON);
#define SDAC_EN digitalWrite(SDAC_SEL_n , OFF);
#define SDAC_DIS digitalWrite(SDAC_SEL_n , ON);
#define TP1_ON  digitalWrite(TPT1_PIN, ON)
#define TP1_OFF  digitalWrite(TPT1_PIN, OFF)
#define TP1_TGL  digitalWrite(TPT1_PIN, digitalRead(TPT1_PIN))
#define TP2_ON  digitalWrite(SEVEN_PIN, ON)
#define TP2_OFF  digitalWrite(SEVEN_PIN, OFF)
#define TP2_TGL  digitalWrite(SEVEN_PIN, digitalRead(TPT1_PIN))

#endif

// selection of sensor-channels is via the ADC multiplexer
//   this code will compile to about as fast as it can be done
#define READ_SNSR_0  ADMUX = (ADMUX & 0xf0) |0x02;    // sensor A is AIN2
#define READ_SNSR_1  ADMUX = (ADMUX & 0xf0) |0x01;    // sensor B is AIN1
#define READ_SNSR_2  ADMUX = (ADMUX & 0xf0) |0x00;    // sensor C is AIN0
#define READ_TEMP    ADMUX = (ADMUX & 0xf0) |0x08;    // #?????
// more I/O macros
#define ADC_ON            ADCSRA |=  0x80;        // ADC-converter ON
#define ADC_OFF           ADCSRA &= ~(0x80);      //   ... and OFF
#define MUX_2_COMP        ADCSRB |=  0x40;        // connect analog-mux to COMPARATOR
#define MUX_2_ADC         ADCSRB &= ~0x40;        //  ...  to ADC
#define ACOMP_ENABLE      ADC_OFF; MUX_2_COMP;    //  enable analog-comparator and connect to input-mux
#define ACOMP_INT_ENABLE  ACSR |= 8;              // enable AComp interrupt
#define ACOMP_INT_DISABLE ACSR &= ~8;             // disable AComp interrupt


#define DISP_DECAY   45   // sets decay-rate of peak-detector display LED
#define LAST_STEP    9    // number of 500-uSec steps per conversion
// channel-numbers as symbols 
#define CH_A  0           // so we can use a meaningful value .....
#define CH_B  1           //  ... for comparisons to variable 'ChNum'
#define CH_C  2
// channel-numbers as bit-masks.. used in noise-setup
#define CHM_A 1
#define CHM_B 2
#define CHM_C 4
#define CH_ALL (CHM_A + CHM_B + CHM_C)

#define SOME_INFO   1 // flags for information-display
#define ALL_INFO    2
#define SIGNED      1 // flags for ASCII-converter
#define UNSIGNED    0

// variables
volatile uint16_t LongTime = 0;
volatile uint8_t  ChNum = CH_A;         // channel #(0,1 or 2) used in current measurement
volatile uint8_t  Step = 0;             // measurement-cycle step # (0-9(?))
uint8_t  Latest_SCI[] = { 'A','=','1','3','5','7','9','0',10};  // buffer for sending values serially
volatile uint16_t Latest_SPI = 12345;   // for assembling SPI-outputs
uint16_t Latest_Temp;                   // most-recently-measured temperature
volatile uint16_t Peak_Detect = 0;      // for driving LED

volatile uint16_t FirstTrip;      // time at crossing of lower threshold
volatile uint16_t SecondTrip;     //  .. ditto upper threshold
volatile uint8_t  AC_Edge = 0;    // sequence-# of comparator state-machine
volatile uint16_t SCount = 0;     // counter for fixed-length outputs (3000 values)
volatile uint8_t  ReStart = CH_ALL;    // flag for 'running setup'
volatile uint8_t  ARate = 1;     // Averaging rate (1,2,4 or 8) >>Loaded from EEPROM at reset<<<
// these vars are used locally/immediately by the reader
//   and thus not tied to a particular sensor-channel 
volatile uint16_t Too_High;      // for noise-rejection
volatile uint16_t Too_Low;       // ditto
volatile uint16_t NoWayLow;      // could not possibly be this low 
volatile uint16_t NoWayHi;       // could not possibly be this high 
volatile uint16_t Max_Change;    // largest allowed change from one reading to the enxt
volatile uint16_t Tared;          // time value with zero-adjust applied
// here are variables that are tied to a particular sensor-channel ( 1 of 3)
//   and whose values must persist during interleaved channel-reading
volatile struct ChInfo {
  uint32_t  Acc32;      // Noise-setup/averaging ACCUMULATOR
  uint8_t   RST_Cntr;   // Noise-setup COUNTER
  uint16_t  SnsrVal;    // most-recently acquired time    ....   for this channel
  uint16_t  Filtered;   // most-recently accepted value
  uint16_t  Check;      // value from noise-trap
  uint16_t  ABuf[16];   // averaging    BUFFER
  uint16_t  D_Ave;      // averaging   AVERAGE
  uint16_t  Basis;      // unstretched average.. important #
  uint16_t  Shown;      // value to be sent out
  uint16_t  Zero;       // tare-value for this channel >>Loaded from EEPROM at reset<<<
} CHinfo[3];
// a shorthand name for 'Current Channel's Information Array'
#define CCH CHinfo[ChNum]

volatile uint8_t ChSeq_Num = 0;    // [0:6]  points to ChSeq[] major-index >>Loaded from EEPROM at reset<<<
volatile uint8_t ChSeq_Ptr = 0;    // [0:5]  indicates current channel

// ---- flag-bytes that control I/O operation -----
//  one is changed asynchronously by the command-interpreter,
//  and  the other is synchronously made to reflect the first
volatile uint8_t Flag_Asy = 0;     // >>Loaded from EEPROM at reset<<<
volatile uint8_t Flag_Syn = 0;
// flags
#define SERIAL_F    1   // if set: enable serial (UART) output
#define DAC_F       2   // if set: enable Quad-DAC output
#define NAKED       4   // if set: data has NO HEADER-IDs 
#define HELLO       8   // if set: show splash-screen
// output-control macros
#define UART_ON       Flag_Asy |= SERIAL_F;   // macro 'turn UART on'
#define SDAC_ON       Flag_Asy |= DAC_F;      // ... 'turn DAC on'
#define HEADERS_ON    Flag_Asy &= ~NAKED;     // ... 'show headers' (A,B or C at front of values)
#define HELLO_ON      Flag_Asy |= HELLO;     // ... 'show greeting-screen'
#define UART_OFF      Flag_Asy &= ~SERIAL_F;  // macro 'turn UART off'
#define SDAC_OFF      Flag_Asy &= ~DAC_F;     // ... 'turn DAC off'
#define HEADERS_OFF   Flag_Asy |=  NAKED;     // ... 'no headers' ... just values and commas
#define HELLO_OFF     Flag_Asy &= ~HELLO;     // ... 'no greeting-screen'
// flag-test macros
#define UART_IS_ON        (  Flag_Syn & SERIAL_F  )   // macro: 'test whether UART is on'
#define SDAC_IS_ON        (  Flag_Syn & DAC_F     )
#define HEADERS_ARE_ON    ( !(Flag_Syn & NAKED  ) )
#define SAYING_HELLO      (  Flag_Syn & HELLO     )

// ===== values (constants) in Flash-memory ===========
const uint8_t RsMask[] = { CHM_A, CHM_B , CHM_C} ;  // map channel-# to a bit-mask
// pre-built headers for sending to serial-DAC
const uint16_t DacHdr[] = { 0x1000, 0x5000, 0x9000, 0xD000};
// --- channel-sequences ----------
//  sequences are seven in number,  each six bytes long ---------
//   (7 elements because we don't have a 'read no channels' sequence)
//   (6 bytes    because sequences are 1,2 or 3 channels long, and 6=LCM[1,2,3] )
const uint8_t Ch_Vals[] =
{   CH_A, CH_A, CH_A, CH_A, CH_A, CH_A,  // 1: All A
    CH_B, CH_B, CH_B, CH_B, CH_B, CH_B,  // 2: All B
    CH_A, CH_B, CH_A, CH_B, CH_A, CH_B,  // 3: A,B
    CH_C, CH_C, CH_C, CH_C, CH_C, CH_C,  // 4: All C
    CH_A, CH_C, CH_A, CH_C, CH_A, CH_C,  // 5: A,C
    CH_B, CH_C, CH_B, CH_C, CH_B, CH_C,  // 6: B,C
    CH_A, CH_B, CH_C, CH_A, CH_B, CH_C   // 7: A,B,C  
    };
//  ^--the first column above is the channel-# at which the sequence output
//   should have a line-feed instead of a comma for pretty .CSV-output
// here's a lookup table of the first channel in each sequence...
//   needed for initialization
const uint8_t FirstChannel[]={0, CH_A, CH_B, CH_A, CH_C, CH_A, CH_B, CH_A };

// ======================== setup function ===========================
// ------------- this is called once after reset -------------------
void setup() {
  //   ----initialize digital output-pins----
  pinMode(LED_ACT, OUTPUT);      // activity LED
  pinMode(CHMUX_0, OUTPUT);      // channel-select...
  pinMode(CHMUX_1, OUTPUT);      // ... multiplexer
  pinMode(SDAC_SEL_n, OUTPUT);   // monitor-DAC select-
  pinMode(HI_LO_SEL, OUTPUT);    // reference-select
  pinMode(TPT1_PIN, OUTPUT);     // test-point
  pinMode(SEVEN_PIN, OUTPUT);    //  also test-point 2
  pinMode(A4_PIN, OUTPUT);    // also ditto 3
  //  ... and input-pins .....
  pinMode(CREF_IN , INPUT);     // analog-comparator's 'plus'-pin is an ANALOG-input...
  pinMode(SNSR_CH0, INPUT);     // .. as are the three sensor-inputs
  pinMode(SNSR_CH1, INPUT);
  pinMode(SNSR_CH2, INPUT);

  // Sensor-reading is performed in a series of 10 steps, each taking 500 uSec,
  //  with hardware-reading interleaved with value-display
  //  Each reading is performed and displayed in 5mSec. This works if the
  //   serial-output routine can be can be driven at 1 character every 500 uSec
  // Minimum baud-rate is thus (@ 1 start/8 bits/one stop) 20 bits/mSec or 20Kbaud
  //  Nearest standard rate is 38.4 Kbaud  
  Serial.begin(38400); 
          
  // SPI is used to display to the quad serial-DAC
  SPI.begin();
  SPCR = 0b01011000;              // enable SPI as master, protocol as  sample-rising-edge
  ToSDAC( 0b0010000000000000 );   // clear all DACs

  // initialize the hardware
  ALL_CH_OFF;                  
  LOW_REF_ON;
  SDAC_DIS;
  
  //---------- load the saved setup -----------
  //    read and set the display-setup
  cli();          // disable interrupts during EEPROM-read
  Flag_Asy = EEPROM.read(100);
  Flag_Syn = Flag_Asy;
  //    get last-channel spec and initialize channel-scan
  ChSeq_Num = EEPROM.read(101);
  if ( ChSeq_Num > 7)
      ChSeq_Num = 0;       // (so that this is correct from an uninitialized EEPROM)
  ChSeq_Ptr = 0;
  // must have this ready ahead of time
  ChNum = FirstChannel[ChSeq_Num];     
  //    read/set the zero spec for each channel
  for (uint8_t j = 0; j < 3; j++) {
    CHinfo[j].Zero = 256 * EEPROM.read(102 + j * 2) + EEPROM.read(102 + j * 2 + 1);
    }
 for (uint8_t j = 0; j < 3; j++) {
    CHinfo[j].D_Ave     = 0xFFFF;   // salt the averages
    CHinfo[j].RST_Cntr  = 0;       // make them all eligible for re-calibration
    }
  ReStart = CH_ALL;             // signal to re-start limit-factors on all channels
  ARate  = EEPROM.read(140);  
  // start out silent;
  UART_OFF;
  SDAC_ON;  

  if (SAYING_HELLO) 
      ShowInfo( ALL_INFO  );
  
  //--------------------------------------------------------------
  //   - - - - - - - - - - start a timer - - - - - - - - - - - -
  // Sensor-reading is performed in a series of 10 steps, each taking 500 uSec
  // Well, why not use the native Arduino-environment timer? (Timer0)
  //  That runs at 1.024KHz, (which could be changed) and we don't know exactly 
  //  how long it takes to do whatever it does
  // BUT, the Arduino timer cannot be eliminated or ignored, since it handles
  //   handles the serial-receive/-transmit  functions
  // So, the program-timer will temporarily disable the Arduino timer
  //  to avoid disruption, then re-enable it for aproxiamtely 75% of every  5 mSec
  //  interval, giving time to catch any incoming serial-commands, which are infrequent and brief
  
  // we will use TIMER2 ...
  // set timer2 interrupt at 2kHz
  TCCR2A = 0;   // TCCR2A register = 0 (only concerned with PWM except as noted below)
  TCCR2B = 0;   // same for TCCR2B
  TCNT2  = 0;   // initialize counter value to 0
  // set compare match register for 8khz increments @ Arduino clock of 16MHz
  OCR2A = 125;  //  // 124;// = (16*10^6) / (8000*8) - 1 (must be <256)
  // turn on CTC mode
  TCCR2A |= (1 << WGM21);
  // Set bits for 64 prescaler
  TCCR2B |= (1 << CS02); ///(1 << CS01) | (1 << CS00);
  
  // enable timer compare interrupt
  TIMSK2 |= (1 << OCIE2A);
  
  // the analog-comparator will be used to detect the voltage on the capacitive elements
  // --- analog comparator setup ------
  // 7 ACD   +disable              0  turn it on, duh
  // 6 ACBG  +use bandgap          0  off, but not always in this program
  // 5 ACO   +enable output        1  output will trigger the IC
  // 4 ACI   +interrupt occurred   0  (read-only)
  // 3 ACIE  +int-enable           0  to be set when needed
  // 2 ACIC  + trigInputCapture    1  ON
  // 1:0 ACIS[1:0]  set int-edge   10 falling-edge trigger
  ACSR = 0b00101110;

  // connect analog-input mux to the comparator
  MUX_2_ADC;
  // --- timer input Capture ------
  TCCR1A = 0;      // no output-compares, no PWM
  //  7  ICNC1      1  noise-cancellation on
  //  6  ICES1      0  edge-select NEG-EDGE
  //  5   --
  //4:3  WGM[3:2]  xx  no PWM used

  TCCR1B = 0b10000010;    //  use 2MHz counter
    
  TIMSK1 = 0;       //  we read the IC-value, but no interrupt is generated 

  Step = 0;         // initialize the measurement state-machine

  sei();

  WaitFor( 50); 

}


/////////////////////////////////////////////////////////////////////////////////
// The timing-interrupt's ISR is where almost everything happens
//  It is broken into 10 steps at 500 uSec intervals, giving
//    a read-cycle time of 5mSec, a 200Hz update rate
// (Note: the steps are at 500uSec intervals, but, as of 11Jan2016, a maximum of
//   65 of those microseconds are used. The total time required is 
//    largely that of the charging-circuit. This could be reduced by using a 
//    higher charge-current (now 1 microamp.) Doing so would also give greater noise-immunity, 
//    but would require a faster time-reference. And so on: you went to college.)

ISR(TIMER2_COMPA_vect) {
//  TP1_ON; //<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
  
  // --- tick the delay-counter------
  if (LongTime) LongTime--;

  //------------- step-number checker ----------------
  if (Step > LAST_STEP) Step = 0;

  //=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/
  if (Step == 0) {
    ACOMP_ENABLE;      // turn ADC off, connect mux to comparator

    // send first byte of UART-value (from previous reading-cycle)
    if (UART_IS_ON) Serial.write(Latest_SCI[0]);

    // enable analog-mux sensor-input for this go-around
    if (ChNum == CH_A ) { READ_SNSR_0;      EN_CH0; }
    if (ChNum == CH_B ) { READ_SNSR_1;      EN_CH1; }
    if (ChNum == CH_C ) { READ_SNSR_2;      EN_CH2; }
    
    // initialize result to the most-recently-accepted value
    CCH.SnsrVal = CCH.Filtered;  

    AC_Edge = 1;            // start analog-compare interrupt state machine
    TCCR0B = 0;             // disable Arduino interrupt by removing its clock!
    TCNT1 = (uint16_t)0;    // reset capture-time
    ACOMP_INT_ENABLE;       // enable Analog-Compare interrupt
  }
  //=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/
  //  steps 1,2,3,4 just send a character each while waiting on charge-circuit
  if (Step == 1) {
    if (UART_IS_ON)   Serial.write(Latest_SCI[1]);  // sign-character
  }
  //=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/
  if (Step == 2) {
    if (UART_IS_ON) Serial.write( Latest_SCI[2] );
  }
  //=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/
  if (Step == 3) {
    if (UART_IS_ON) Serial.write(Latest_SCI[3]);

  }
  //=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/
  if (Step == 4) {
    if (UART_IS_ON) Serial.write(Latest_SCI[4]);
    // so adjust the average-buffer so next sample always goes in .ABuf[0]
    //  (move every sample downstream by one slot)
    for (uint8_t ictr = 15; ictr !=0 ; ictr--)
        CCH.ABuf[ictr] = CCH.ABuf[ictr-1];
  }

  //=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/
  // by step 5, the charge-circuit *SHOULD* be done... 
  //  a 125mm sensor at 100%-strain/1uAmp just barely makes it
  if (Step == 5) {
    // send another digit of previous reading
    if (UART_IS_ON) Serial.write(Latest_SCI[5]);
    ACOMP_INT_DISABLE;      // make sure of this
    TCCR0B = 3;             // restart Arduino interrupt!
    LOW_REF_ON;             // measurement is done now, so reset reference-input
    ALL_CH_OFF;             // hold all charge-circuits clamped
    // set maximum-allowed deviation for later
    if(CCH.Basis != 0)  Max_Change =    (CCH.Basis / 8) + (CCH.Basis / 8) ;    
      else              Max_Change = 400;
  }

  //=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/
  if (Step == 6) {
    // sneak in another digit
    if (UART_IS_ON) Serial.write(Latest_SCI[6]);

    // is noise-setup REQUESTED for this channel? 
    if ( ReStart & RsMask[ChNum] ) { 
        CCH.RST_Cntr = 64;              // if that's the case, set this channel's counter
        CCH.Acc32 = 0;                  // initialize the accumulator
        ReStart &=  ~(RsMask[ChNum]);   // disable request

        }
    // =========== noise-rejection setup ==============
    // ... it ASSUmed that the sensor is not moving for this!
    //  Accumulate 64 samples, making sure that any egregiously bad values are not counted 
    //  Once a good set is obtained, find its average and put that in the whole average-buffer
    if ( CCH.RST_Cntr != 0 ){
        // for this, we average a large number of readings of an
        //  UNSTRAINED sensor, so we accept in this range:
        Too_Low  = 400;     // must be > (about  .5x(minimum for a short sensor))
        Too_High = 5000;    //  must be < (about 2.5x(min of a long sensor))
        // check the latest aagainst those limits
        if ( (CCH.SnsrVal > Too_Low) && (CCH.SnsrVal < Too_High) ) {
            CCH.RST_Cntr--;                     // tick down the counter for acceptable readings
            CCH.Acc32 += (uint32_t)CCH.SnsrVal;
            CCH.Shown =  CCH.SnsrVal;           // use this value for display down below
            }// end of good-sensor-reading
        // this channel's restart-counter may have hit zero in the previous part...
        // .. if that's so, move starting average to buffer 
        //     and disable restart for this channel
        if (CCH.RST_Cntr == 0){
            CCH.D_Ave = CCH.Acc32/64;   // div acc by 64
            for (uint8_t ictr = 0; ictr<8; ictr++)
                CCH.ABuf[ictr] = CCH.D_Ave;    // put that value in all eight averaging slots
            CCH.Filtered   =  CCH.D_Ave;    // keep it as noise-tracking value
            CCH.Check      =  CCH.D_Ave;
            CCH.Shown      =  CCH.D_Ave;
            CCH.Basis      =  CCH.D_Ave;                  
            }// end 'last of 64 good-readings-obtained for this channel'
       }// end of noise-setup sample(s)
    
    // =====samples subsequent to noise-setup  ======================
    // once the noise-filter has been initialized for the current channel, 
    //   subsequent readings are done normally
    if ( ((ReStart & RsMask[ChNum])== 0)&&(CCH.RST_Cntr == 0) ) {
        // set limits on the change of this one versus the last
        Too_Low   =  CCH.Filtered - CCH.Basis/2;  //350; 
        Too_High  =  CCH.Filtered + CCH.Basis/2;  //350; 
        // also two 'totally wrong' values  
        NoWayLow  =  CCH.Basis - CCH.Basis/8; //100; 
        NoWayHi   =  CCH.Basis * 4;     
        // ------ noise filter -----------
        CCH.Check = CCH.SnsrVal;          // first, assume that the new input value is OK
        ///  compare input to limits, replacing bad ones with a value that preserves
        //    the SENSE of the change while limiting its SIZE     
        if ( CCH.SnsrVal > Too_High  )    CCH.Check = CCH.Filtered + Max_Change;  
        if ( CCH.SnsrVal < Too_Low   )    CCH.Check = CCH.Filtered - Max_Change;
        // if value is outlandishly wrong, use previous good-value     
        if ( CCH.SnsrVal < NoWayLow  )    CCH.Check = CCH.Filtered; 
        if ( CCH.SnsrVal > NoWayHi   )    CCH.Check = CCH.Filtered; 
        // --- find the running average: (sum-of-N-samples)/N  for N = [1,2,4,8]
        CCH.ABuf[0]    = CCH.Check;       // put accepted value into previously-adjusted averaging buffer
        CCH.Acc32      = 0;               // init average-acc to zero
        switch(ARate){                    // based on averaging-rate...
        case 2:   CCH.D_Ave = ( CCH.ABuf[0] + CCH.ABuf[1] )/2;  
                  break;
        case 4:   CCH.D_Ave  = CCH.ABuf[0] + CCH.ABuf[1] + CCH.ABuf[2] + CCH.ABuf[3]; 
                  CCH.D_Ave  = CCH.D_Ave / 4;  
                  break;
        case 8:   for (uint8_t ictr = 0; ictr<8; ictr++)
                      CCH.Acc32   += (uint32_t)CCH.ABuf[ictr];
                  CCH.D_Ave =  CCH.Acc32 / 8;
                  break;
        default:  CCH.D_Ave = CCH.ABuf[0]; 
                  ARate = 1;            // illegal values (and 1) are treated-as/changed-to 1.. i.e. no averaging) 
                  break;      
         }// end averaging-switch
      CCH.Filtered =  CCH.Check;      // update this
      CCH.Shown    =  CCH.D_Ave;      //  averaged value is the output (might be same as input, if 'ARate' = 1)------   
    } // end of 'samples after noise-setup'

    //===== reading has been bounds-checked and maybe averaged ======================
    //  so we can assemble the next output..
    //----- remove the zero-reference from the time --------
    Tared = CCH.Shown  - CCH.Zero;
    // ---- assemble serial-output string --------
    Latest_SCI[0] = ' ';              // assume no headers!
    if (HEADERS_ARE_ON )  Latest_SCI[0] = 'A' + ChNum;      //if headers are enabled, put in the channel number
    // --- put number in ASCII into buffer following (maybe)header and sign------
    ToASCII16( Tared, &Latest_SCI[1] , SIGNED );  
   
    // ===== make the Activity LED show when any sensor is extended=======
    analogWrite(LED_ACT, Peak_Detect/64);     // PWM the LED to current value
    if (Peak_Detect < abs(Tared) )        // keep track of peaks
            Peak_Detect = abs(Tared);
    if (Peak_Detect > DISP_DECAY)                   // make the peak-detection 'leaky' 
            Peak_Detect -= DISP_DECAY;
    //======= calculate 'Latest_SPI' value and send it to the DAC =======
    //   because time-values are ~16,000 max, divide
    //   them by 4 for the 12-bit [0:4095] DAC
    //  don't let values over/under-run the DAC-range
    if (Latest_SPI > 20000) Latest_SPI = 0;       // some are negative #'s
    if (Latest_SPI > 4095)  Latest_SPI = 4095;
    //   add the channel-specific header to the value and send it
    Latest_SPI = DacHdr[ChNum] + Latest_SPI;
    if (SDAC_IS_ON)   ToSDAC( Latest_SPI );

    // make sure the comparator doesn't fire again
    ACOMP_INT_DISABLE;


    // ====== read pot using 5V ref, (which was set earlier) ========
    analogRead( POT_IN );
    // change analog-ref to 1.1V
    analogReference(INTERNAL);



  }

  //=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/
  if (Step == 7) {
    // ---- send final UART-char (line-feed or comma) -----
    //   every sequence has a 'last reading' .. see if this it
    if (UART_IS_ON) { 
        if (ChNum == Ch_Vals[ChSeq_Num*7])    // look up the number to be considered 'last'
            Serial.write(10);                 // send a line-feed for thjat one
        else  
            Serial.write(',');                // send a comma for all others
        }
    
    // ===== read/send temperature ======
    Latest_Temp = analogRead(8);            // read internal temp-sensor )

    ToSDAC( DacHdr[3] + (Latest_Temp * 4)  ); // change 8-bit value to 10-bits and send
    ADC_OFF;    //                          // turn off ADC so that comparator works
    MUX_2_COMP;                             // (re-)connect ADC-mux to comparator
  }

  //=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/=/
  if (Step == LAST_STEP) {
    //----- things to do at the end of a measurement ----

    // .....look for and implement changes to display-modes ....
    // display-controls only change at the
    //   END OF THE FIRST sequence-step, because displayed values are
    //   from the PREVIOUS step, and we want the sequence to have
    //   a length of six, but with the same number of A,B or C values
    if ( ChSeq_Ptr == 0){ 
      
      // there may be triggered-sample-set underway
      if ( SCount)  {      // if triggered-send has counts left...
          SCount--;           // then tick down
          UART_ON;            // assume we're not done
          if (!SCount)        // but if that WAS the end,  
                UART_OFF;         // stop sending 
          } // end of 'triggered-send underway'
        
      // transfer Asynchronous flag-byte to its Synchronous bro  
        if (Flag_Asy != Flag_Syn)
          Flag_Syn = Flag_Asy;
      } // end of 'end first output of sequence '
    
    //... increment and bounds-check the channel number pointer....
    ChSeq_Ptr++;  if (ChSeq_Ptr > 5)
      ChSeq_Ptr = 0;

    //.... with sequence-pointer set, now set new channel-number ....
    ChNum = Ch_Vals[ 6 * ChSeq_Num + ChSeq_Ptr];

    }// end last-step

  Step++;   // increment step-pointer at end of each step

} // end of ~500uSec Timer ISR



////////////////////////////////////////////////////////////////////////////////////////////////
// The analog comparator interrupt is called three times per measurement:
//   1 Sensor-voltage increases past the lower threshold (.5 volts)
//   2 Glitch caused by reference-voltage suddenly changing from LOW to HIGH
//   3 Sensor-voltage increases past the upper-threshold (2.5 volts)
// In cases 1 and 3, the Input Capture unit's recorded time is read and used to
//   determine the net charging time and, by inference, the sensor elongation
ISR(ANALOG_COMP_vect)
{ switch (AC_Edge) {
    // first Analog-Comparator interrupt is when input crosses lower-threshold
    case 1:
      FirstTrip = ICR1;   // save time-stamp  as start-time
      TP2_ON;             // <<< scope-output <<<<<
      AC_Edge = 2;        // move state-machine to second interrupt
      HIGH_REF_ON;        // switch comparator input to higher voltage

      break;
    // second AC-interrupt is IGNORED, as it is caused by reference-input change
    case 2:
      SecondTrip = ICR1;  //  save time-stamp as finish time JUST IN CASE
      CCH.SnsrVal = SecondTrip - FirstTrip;
      AC_Edge = 3;        // move state-machine to next
      TP2_OFF;
      break;
    // third interrupt is when input-voltage crosses upper threshold
    case 3:    /// 3:
      SecondTrip = ICR1;  //  save time-stamp as finish time;
      TP2_ON;             // <<< scope-output <<<<<
      CCH.SnsrVal = SecondTrip - FirstTrip; // note DIFFERENCE
      AC_Edge = 0;        // reset state-machine
      ACOMP_INT_DISABLE;  // disable this interrupt-routine

      TCCR0B = 3;     // restart Arduino interrupt!

      TP2_OFF;


      break;
  } // end 'switch-case'

} // end Analog Comparator ISR


////////////////////////////////////////////////////////////////////////////////////////////////
//   The loop function just sits waiting for command-inputs, little knowing how much activity 
//  goes on in the interrupt routines
//-----------------------------------------------------
// Control Commands
//  Sensor-reader demo operation is controlled from the UART-input at
//  38400 baud using a simple ASCII protocol
//  All commands are case-insensitive, and no spaces are allowed.
//  Any incorrect character will terminate a command.
//
//   Y  ENABLE display of sensor-data via UART at 38400 baud
//
//   X  DISABLE display via UART  (sensor-reading continues, however)
//
//   M<n>    Set measurement sequence:
//          <n> = [1,2,3,4,5,6,7]
//           M1 Measure and display data from sensor A only  AAAAAA
//           M2 Measure and display data from sensor B only  BBBBBB
//           M3 Measure and display data from  A and B       ABABAB
//           M4 Measure and display data from sensor C only  CCCCCC
//           M5 Measure and display data from  A and C       CACACA 
//           M6 Measure and display data from  B and C       BCBCBC
//           M6 Measure and display data from  A, B and C    ABCBAC
//           Measuring a single-sensor devotes every measurement-cycle to it, giving the
//             highest possible update-rate. A sensor-reading is taken and displayed every 5 milliseconds, so
//             the update rate is minimum of 1/(3*.005) or 67Hz for three channels and
//             a maximum of 200 Hz for one channel
//
//   Z<n><a>  Change the Zero-value for channel <n>
//           <n> = [A,B,A]
//           <a> = [A,C]
//         The time-values from sensor-measurement can be modified for display by addition
//           of a 'Zero' (offset) value for each sensor. Raw measured values are
//         in units of time, and range from #?#?#?#? to ?#?#?#?.
//         At the current time, the two possible actions are
//         'C'    CLEAR the Zero of the specified channel (0,1,2)
//         'A'    use the AVERAGE value for the Zero of the specified channel (0,1,2)
//                 (average of the past 16#?#? readings)
//         Examples:  "ZAC" = 'clear zero-value for channel A'
//
//   E  This command saves the current values of the above(display, sequence, zeros)
//        to non-volatile memory. At the next power-on of the demo, controls will be in the configuration
///       saved by the most-recent 'E'-command.
//        As delivered, the board will display all three channels via UART and DAC, with all Zeros cleared,
//
//   F  shows zero-values and current averages, after doing a "Q"
//   G  Shows help-menu

//--------------------------------------------------------------------------
// ============== loop function ============================================
void loop() {
  // the 'loop' task watches for serial-inputs that change the operation
  uint8_t Chr_1st;
  uint8_t Chr_2nd;
  uint8_t Chr_3rd;
  uint8_t Work_Chnl;

  // see if a character has  appeared on the serial-input
  //  thereafter, operation is different for each command
  if (Serial.available() ) {
    Chr_1st = toupper(Serial.read());      // make an upper-case version of input
    // ----- single-character commands -------


  switch (Chr_1st){
      case 'X':   UART_OFF;      SCount  =  1; break;     // stop sending ( also properly close a sequence-send)
      case 'Y':   UART_ON;       ReStart = CH_ALL; break;    // start sending
      case '!':   SCount = 501;  ReStart = CH_ALL; break;     //show 6*500 = 3000 values*(meas-length)
      case 'V':   HEADERS_ON;; break;     // show a header on each measurement
      case 'W':   HEADERS_OFF; break;     //    ... or don't
      case 'S':   HELLO_ON; break;     // show information at start-up
      case 'U':   HELLO_OFF; break;     //   ... or don't
      case 'H':   ARate = 1; break;     // no averaging
      case 'J':   ARate = 2; break;     // average two
      case 'K':   ARate = 4; break;      // average four
      case 'L':   ARate = 8; break;      // average eight

     case 'F':       // ...... 'F' and 'G' display information  .................
      case 'G':
          // if UART-display is running, turn it off, and wait til it's stopped
          if (UART_IS_ON)  {  UART_OFF; while( UART_IS_ON){}; }
          //  then display debugging stuff
          if (Chr_1st == 'F') ShowInfo( ALL_INFO  );
          if (Chr_1st == 'G') ShowInfo( SOME_INFO  );
          break;     // end debugging-display
 
      case 'M':            // .... measurement-sequence 'M' can be 1,2,3,4,5,6 or 7 ...
          Chr_2nd = GetCharIn();        // wait for another char
          if ( (Chr_2nd >= '1') && ( Chr_2nd <= '7'))
            ChSeq_Num = Chr_2nd - '1' ;     //  .. NOT -'0'... turns 1-7 into 0-6
            ReStart = CH_ALL;
          break;    // end set measurement-length
 
      case 'Z':               // .... zero control: 'A'verage or 'C'lear ..........
          Chr_2nd = GetCharIn() & 0xDF;      // get channel-# ( 'A',B','C' )
          Work_Chnl = 3;              // assume it's wrong (there is no channel > C )
          if ( (Chr_2nd >= 'A') && ( Chr_2nd <= 'C'))
                Work_Chnl = Chr_2nd - 'A'; // resolve 'A',B','C' into 0,1,2
          // get another upper-case char ( 'A' or 'C')
          Chr_3rd = (GetCharIn() & 0xDF);
          // 'C' means 'clear the specified channel's zero'
          if (( Chr_3rd == 'C') && (Work_Chnl< 3) )  
                  CHinfo[Work_Chnl].Zero = 0;
          // 'A' means 'use current-average as channel's zero'      
          if (( Chr_3rd == 'A') && (Work_Chnl < 3) ) 
                  CHinfo[Work_Chnl].Zero = CHinfo[Work_Chnl].D_Ave; 
          break;      // end zero-set
 

      case 'E':         // .... save to EEPROM ...           
          cli();
          analogWrite(LED_ACT, 255); 
          EEPROM.update(100, Flag_Asy);
          EEPROM.update(101, ChSeq_Num);
          EEPROM.update(140, ARate);
          for (uint8_t j = 0; j < 3; j++) {
            EEPROM.update(102 + j * 2, CHinfo[j].Zero >> 8); EEPROM.update(102 + j * 2 + 1, CHinfo[j].Zero & 255);
            }
          sei();
          //delay(900);
          WaitFor( 900);  //
          break;    // end of 'save to EEPROM'

  }// end of switch-on-first-input

  }// end 'input character ready'
} // end 'loop'

////////////////////////////////////////////////////////////////////////////////////////////////
//==== send a WORD-value to the serial-DAC======
//    word must be formatted for desired function
void ToSDAC( uint16_t SPI_word) {
  SDAC_EN;
  SPDR = highByte(SPI_word);
  while ( !(SPSR & 128));
  SPDR = lowByte(SPI_word);
  while ( !(SPSR & 128));
  SDAC_DIS;
}

////////////////////////////////////////////////////////////////////////////////////////////////
// ============ convert a uint16_t to ASCII-decimal =================
//  signed or unsigned, although only signed used here
void ToASCII16( uint16_t value, uint8_t * ASCII_out, uint8_t mode) {
  ASCII_out[0] = ' ';        // assume unsigned (first character = blank)
  if (mode == SIGNED){     // If it is signed
    ASCII_out[0] = '+';     // assume positive
    if (value > 32767) {    // but if it IS negative
      ASCII_out[0] = '-';          // fix to suit
      value = ~value;
      value += 1;
      } // end 'actually negative'
    } // end 'actually signed'
  // now convert it    
  uint16_t DecDgt;
  DecDgt = 10000;  ASCII_out[1] = '0';
  while (value >= DecDgt) {
    value -= DecDgt;
    ASCII_out[1]++;
  }
  DecDgt = 1000;   ASCII_out[2] = '0';
  while (value >= DecDgt) {
    value -= DecDgt;
    ASCII_out[2]++;
  }
  DecDgt = 100;    ASCII_out[3] = '0';
  while (value >= DecDgt) {
    value -= DecDgt;
    ASCII_out[3]++;
  }
  DecDgt = 10;     ASCII_out[4] = '0';
  while (value >= DecDgt) {
    value -= DecDgt;
    ASCII_out[4]++;
  }
  ASCII_out[5] =  value + '0';
}

////////////////////////////////////////////////////////////////////////////////////////////////
//  a delay, in units of interrupts*2 (400 uSec or so)
void WaitFor( uint16_t ThisLong){
  cli();
  LongTime = ThisLong*2;
  sei();
  while ( LongTime == 0){}
}

////////////////////////////////////////////////////////////////////////////////////////////////
//=========  print out some information about setup ===========
void ShowInfo( uint8_t HowMuch) {
  uint8_t x;
  Serial.println(" ");
  Serial.println("Parker Sensor Demo V2.01  2016");
  Serial.println( __DATE__ "/" __TIME__);
  Serial.print("  Sequence\t"); 
  switch (ChSeq_Num){
    case 0: Serial.println("AAAAAA"); break;
    case 1: Serial.println("BBBBBB"); break;
    case 2: Serial.println("ABABAB"); break;
    case 3: Serial.println("CCCCCC"); break;
    case 4: Serial.println("ACACAC"); break;
    case 5: Serial.println("BCBCBC"); break;
    case 6: Serial.println("ABCABC"); break;
     }
  Serial.print("  Headers\t");
  if (HEADERS_ARE_ON) Serial.println("ON");
  else                Serial.println("OFF");
  Serial.print("  Average\t"); 
  switch (ARate){
    case 2: Serial.println("2x"); break;
    case 4: Serial.println("4x"); break;
    case 8: Serial.println("8x"); break;
    case 1: 
    default: Serial.println("NONE"); ARate = 1;  break;
   }   

  Serial.print("Zeros: "); for (x = 0; x < 3; x++) {
    Serial.print( CHinfo[x].Zero);  Serial.print( "\t");  }  Serial.println(" ");
  Serial.print("Aves:  "); for (x = 0; x < 3; x++)  {
    Serial.print( CHinfo[x].D_Ave); 
    if(ReStart & RsMask[x]) Serial.print( "?"); Serial.print( "\t"); 
    } 
    Serial.println(" ");
if (HowMuch == ALL_INFO){
    Serial.println (F(" \"Y\"\t\"X\"\tSerial-display ON or OFF"));
    Serial.println (F(" \"V\"\t\"W\"\tData-Headers   ON or OFF"));
    Serial.println (F(" \"M1\" AAAAAA\t\"M2\" BBBBBB\t\"M3\" ABABAB\t\"M4\"  CCCCCC"));
    Serial.println (F(" \"M5\" ACACAC\t\"M6\" BCBCBC\t\"M7\" ABCABC"));
    Serial.println (F(" \"ZAA\"\t\"ZAC\"\tSet CHA tare: Average/Clear"));  
    Serial.println (F(" \"ZBA\"\t\"ZBC\"\tSet CHB tare: Average/Clear"));  
    Serial.println (F(" \"ZCA\"\t\"ZCC\"\tSet CHC tare: Average/Clear")); 
    Serial.println (F(" \"E\"\t\twrite setup to EEPROM")); 
    Serial.println (F(" \"F\"\t\tshow this")); 
    Serial.println (F(" \"G\"\t\tshow values only"));
    Serial.println (F(" \"!\"\t\tsend 3000 samples"));
    Serial.println (F(" \"S\"\t\"U\"\tSign-on Message  ON or OFF"));   
    Serial.println (F(" \"H\",\"J\",\"K\",\"L\"\tAveraging: none,2,4,8"));
  }
}

////////////////////////////////////////////////////////////////////////////////////////////////
// ========= wait for and return a serial-input =================
uint8_t GetCharIn(void) {
  while ( ! Serial.available() ) {}
  return Serial.read();
}

////////////////////////////////////////////////////////////////////////////////////////////////
//   print an 8-bit value in hex
void Print8(uint8_t value) {
  Serial.print( (value & 0xf0) >> 4 , HEX);
  Serial.print( (value & 0x0f) , HEX);
}

////////////////////////////////////////////////////////////////////////////////////////////////
//  print a 16-bit value in hex
void Print16(uint16_t value) {
  Print8(highByte(value));
  Print8(lowByte(value));
}



