function [C]=HSSRead(port,sec)
% HSSRead v1.01  Parker Hannifin Corp  03/15/2016 
% 
% Reads line data from Arduino high strain sensor demo board, 
% strips the data of the data header prefix and 
% writes the remaining count data to an excel file, HSSData.xlsx. 
%
% Assumes the following settings are saved to demo board EEPROM:
%   1) sign-on message is set ON
%   2) single channel read sequence (i.e. AAAAAA, BBBBBB, or CCCCCC)
%
% [C]=HSSRead(port,sec)
% port = communication port (e.g. 'COM11')
% sec = aquisition time in number of seconds
% C = data output (in counts)
%
% example:
% >> [C]=HSSRead('COM11',10);

% prepare serial port to Arduino
s1 = serial([port]);
set(s1,'BaudRate',38400);
s1.Terminator = 'LF';
npts=sec*200; % number of points @ 200 samples per second
A=zeros(npts,7); % create sized arrays
B=zeros(npts,5);
C=zeros(npts,1);

fopen(s1); %open serial port

% acquire lines for sign-on message (these lines are discarded)
for i=1:21 % # of lines in sign on message
    D = fgetl(s1);
end

disp(['acquiring']); % visual prompt for identifying start of acquisition
fprintf(s1,'Y'); % start data
% acquire specified number of lines
for n=1:npts 
    A(n,:) = fgetl(s1);
end
fprintf(s1,'X'); % stop data

% close serial port
fclose(s1);
delete(s1)
clear s1

% strip data of header prefix
for n=1:npts
    B(n,1:5) = A(n,3:7); % grab last 5 characters of data
    C(n,1) = str2num(native2unicode(B(n,:))); % convert to unicode
end

figure; plot(C); % plot data
xlswrite('HSSData.xlsx',C,1) % write data to excel
end
